import math
import json
import pymel.core as pm
import maya.OpenMaya as om
import maya.OpenMayaUI as omui
from PySide2 import QtGui, QtCore, QtWidgets
from shiboken2 import getCppPointer, wrapInstance

## ----------------------------------------------------------------------------##
# Scripts for working with fake instances (duplicates) of original model
## ----------------------------------------------------------------------------##
# used when you need lots of same objects in scene, but you dont want to use real referenced duplicates for some reason


# To to:
# Option to make instance copies or not
# Instance name field fix.
# Duplicate

############# UI #############
class InstancerMainWindow(QtWidgets.QDialog):
    def __init__(self, parent=None):
        window = omui.MQtUtil.mainWindow()
        mayaWindow = wrapInstance(int(window), QtWidgets.QMainWindow)
        super(InstancerMainWindow, self).__init__(mayaWindow)
        self.setWindowTitle("Instancer")
        self.setMinimumWidth(100)
        self.setMinimumHeight(250)
        main_layout = QtWidgets.QVBoxLayout(self)
        main_layout.addWidget(InstancerUIWidget())
        self.show()


class InstancerUIWidget(QtWidgets.QWidget):
    # Parent is main Maya Window
    def __init__(self, parent=None):
        print("here")
        QtWidgets.QWidget.__init__(self)
        print("here1")
        self.createUI()
        print("here2")
        self.selection_tracking_scriptjob()

    def createUI(self):
        # Main layout
        mainLayout = QtWidgets.QVBoxLayout()
        self.setLayout(mainLayout)

        lay_instance_name = QtWidgets.QVBoxLayout()
        self.led_instance_name = QtWidgets.QLineEdit()
        self.led_instance_name.setDisabled(True)
        lbl_instance_name = QtWidgets.QLabel("Instance name:")
        lay_instance_name.addWidget(lbl_instance_name)
        lay_instance_name.addWidget(self.led_instance_name)

        lay_original_name = QtWidgets.QVBoxLayout()
        self.led_original_name = QtWidgets.QLineEdit()
        self.led_original_name.setDisabled(True)
        lbl_original_name = QtWidgets.QLabel("Original name:")
        lay_original_name.addWidget(lbl_original_name)
        lay_original_name.addWidget(self.led_original_name)

        mainLayout.addLayout(lay_original_name)
        mainLayout.addLayout(lay_instance_name)

        self.wid_originals_btns = QtWidgets.QWidget()
        # self.wid_originals.setHidden(True)
        lay_original_btns = QtWidgets.QVBoxLayout()
        self.wid_originals_btns.setLayout(lay_original_btns)
        self.wid_originals_btns.setHidden(True)

        btn_update_instances = QtWidgets.QPushButton("Update Instcs")
        btn_update_instances.clicked.connect(self.update_instances)
        lay_original_btns.addWidget(btn_update_instances)

        btn_make_instance = QtWidgets.QPushButton("Make Inst")
        btn_make_instance.clicked.connect(self.make_instance)
        lay_original_btns.addWidget(btn_make_instance)

        mainLayout.addWidget(self.wid_originals_btns)

        self.wid_instance_btns = QtWidgets.QWidget()
        self.wid_instance_btns.setHidden(True)
        lay_instance_btns = QtWidgets.QVBoxLayout()
        self.wid_instance_btns.setLayout(lay_instance_btns)

        self.btn_duplicate_instance = QtWidgets.QPushButton("Duplicate Inst")
        self.btn_duplicate_instance.clicked.connect(self.duplicate_instance)

        self.btn_replace_original = QtWidgets.QPushButton("Replace Orig")
        self.btn_replace_original.clicked.connect(self.replace_original)
        self.btn_select_original = QtWidgets.QPushButton("Select Orig")
        self.btn_select_original.clicked.connect(self.select_original)
        lay_instance_btns.addWidget(self.btn_duplicate_instance)
        lay_instance_btns.addWidget(self.btn_replace_original)
        lay_instance_btns.addWidget(self.btn_select_original)
        mainLayout.addWidget(self.wid_instance_btns)

        self.btn_make_original = QtWidgets.QPushButton("Make Orig.")
        mainLayout.addWidget(self.btn_make_original)
        self.btn_make_original.clicked.connect(make_original)
        self.btn_make_original.setHidden(False)

        self.cbx_export = QtWidgets.QCheckBox("Export")
        self.cbx_export.stateChanged.connect(self.export_state_changed)
        btn_export_originals = QtWidgets.QPushButton("Export Originals")
        btn_export_originals.clicked.connect(self.export_originals)
        btn_export_originals.setHidden(True)
        btn_export_instances = QtWidgets.QPushButton("Export Instances")
        btn_export_instances.clicked.connect(self.export_instances)
        btn_export_instances.setHidden(True)
        btn_update_instances_json = QtWidgets.QPushButton("Update Instances")
        btn_update_instances_json.clicked.connect(self.update_instances_json)
        btn_update_instances_json.setHidden(True)

        lay_export_path = QtWidgets.QVBoxLayout()
        self.led_export_path = QtWidgets.QLineEdit("C:/_Temp/Instancer/")
        btn_change_export_path = QtWidgets.QPushButton("...")
        btn_change_export_path.setMaximumWidth(15)
        btn_change_export_path.clicked.connect(self.change_export_path)
        lay_export_path.addWidget(self.led_export_path)
        lay_export_path.addWidget(btn_change_export_path)
        wid_export_path = QtWidgets.QWidget()
        wid_export_path.setHidden(True)
        wid_export_path.setLayout(lay_export_path)

        mainLayout.addWidget(self.cbx_export)
        mainLayout.addStretch()
        mainLayout.addWidget(wid_export_path)
        mainLayout.addWidget(btn_export_originals)
        mainLayout.addWidget(btn_export_instances)
        mainLayout.addWidget(btn_update_instances_json)

        self.export_widgets = [btn_export_originals,
                               btn_export_instances, btn_update_instances_json, wid_export_path]

    def export_originals(self):
        export_originals(self.led_export_path.text())

    def export_instances(self):
        export_instances(self.led_export_path.text())

    def update_instances_json(self):
        update_instances_json(self.led_export_path.text())

    def change_export_path(self):
        path = QtWidgets.QFileDialog.getExistingDirectory()
        self.led_export_path.setText(path)

    def make_instance(self):
        make_instance(self.instance_parent[0])

    def export_state_changed(self):
        if self.cbx_export.isChecked():
            for widget in self.export_widgets:
                widget.setHidden(False)
            self.setMinimumHeight(500)
            self.setMaximumHeight(500)
        else:
            for widget in self.export_widgets:
                widget.setHidden(True)

            self.setMinimumHeight(250)
            self.setMaximumHeight(250)

    def update_instances(self):
        update_instances(self.instance_parent)

    def duplicate_instance(self):
        duplicate_instance(self.instance_parent[0])

    def replace_original(self):
        replaceOriginal(self.instance_parent[0], self.orig)

    def select_original(self):
        pm.select(self.orig)

    def update_led_fields(self):
        sel = pm.ls(sl=1, l=1)
        if sel:
            sel = sel[0]
            self.instance_parent = find_instance_parent(sel)

            if self.instance_parent[0]:
                self.btn_make_original.setHidden(True)
                if self.instance_parent[1]:
                    self.wid_originals_btns.setHidden(False)
                    self.wid_instance_btns.setHidden(True)

                    self.led_original_name.setText(
                        str(self.instance_parent[0]))
                    self.led_instance_name.setText("")
                else:
                    self.wid_originals_btns.setHidden(True)
                    self.wid_instance_btns.setHidden(False)

                    self.led_instance_name.setText(
                        str(self.instance_parent[0]))

                    self.orig = find_original(self.instance_parent[0])
                    if self.orig:
                        self.led_original_name.setText(str(self.orig))
                    else:
                        self.led_original_name.setText("")
            else:
                self.led_original_name.setText("Not An Instance")
                self.led_instance_name.setText("Not An Instance")
                self.wid_originals_btns.setHidden(True)
                self.wid_instance_btns.setHidden(True)
                self.btn_make_original.setHidden(False)
        else:
            self.led_original_name.setText("Not An Instance")
            self.led_instance_name.setText("Not An Instance")
            self.wid_originals_btns.setHidden(True)
            self.wid_instance_btns.setHidden(True)
            self.btn_make_original.setHidden(False)
            # self.led_original_name.setText(
            #    str(sel)[0:sel.find("_inst")] + "_orig")

    def selection_tracking_scriptjob(self):
        for job in pm.scriptJob(listJobs=True):
            if job.find("update_led_fields") != -1:
                job_num = job.split(":")[0]
                pm.scriptJob(kill=int(job_num), force=True)

        pm.scriptJob(e=["SelectionChanged", self.update_led_fields])


# -------------------------------------------------
# instancer scripts
# -------------------------------------------------
def check_if_control(node):
    if node.getShape():
        node_name = node.split("|")[-1]
        if node_name.find("_ctrl") != -1:
            return True
        else:
            return False
    else:
        return False


def copyRigState(source, target):
    """
    Copyes joint transforms from instance to dupeOrig
    Instance: instance of the object to copy transforms from
    dupeOrig: newly duplicated object to copy transforms to
    """

    old_instance_controls = pm.listRelatives(
        source, c=1, ad=1, type="transform")
    old_instance_controls = [
        x for x in old_instance_controls if check_if_control(x)]
    print(old_instance_controls)
    source_clean_name = source.split("|")[-1]
    target_clean_name = target.split("|")[-1]
    for old_instance_control in old_instance_controls:
        new_instance_control = str(old_instance_control).replace(
            str(source_clean_name), str(target_clean_name))

        if pm.objExists(new_instance_control):
            translate = pm.getAttr(old_instance_control + ".translate")
            rotate = pm.getAttr(old_instance_control + ".rotate")
            scale = pm.getAttr(old_instance_control + ".scale")

            pm.setAttr(new_instance_control + ".translate", translate)
            pm.setAttr(new_instance_control + ".rotate", rotate)
            pm.setAttr(new_instance_control + ".scale", scale)

        else:
            print("No joint " + new_instance_control)


def reset_rig_state(instance):
    """
    Resets rig state of instance
    """
    instance_controls = pm.listRelatives(instance, c=1, ad=1, type="transform")
    instance_controls = [
        x for x in instance_controls if check_if_control(x)]
    for instance_control in instance_controls:
        pm.setAttr(instance_control + ".translate", 0, 0, 0)
        pm.setAttr(instance_control + ".rotate", 0, 0, 0)
        pm.setAttr(instance_control + ".scale", 1, 1, 1)

    print("--reset_rig_state----out----")


def jWalkDown(instance, dupeOrig):
    # Goes down joint the hierarchy and copyes transforms from old joints of instance to a new ones
    # initial function
    instname = instance.split("|")[-1]
    dupeOrigName = dupeOrig.split("|")[-1]
    # for every child of instance---
    for obj in pm.listRelatives(instance, c=1, f=1):
        # if its a joint and duplicated original exists
        if obj.split("|")[-1].find("_j_") > 0 and pm.objExists(obj.split("|")[-1].replace(instname, dupeOrigName)):
            # print obj
            # pm.matchTransform(obj.replace(instname, dupeOrigName),obj,rot=1,pos=1)
            newTranslate = pm.getAttr("{0}.translate".format(obj))[0]
            newRotate = pm.getAttr("{0}.rotate".format(obj))[0]
            newScale = pm.getAttr("{0}.scale".format(obj))[0]
            pm.setAttr("{0}.translate".format(obj.replace(instname, dupeOrigName)),
                       newTranslate[0], newTranslate[1], newTranslate[2], type="float3", edit=1)
            pm.setAttr("{0}.rotate".format(obj.replace(instname, dupeOrigName)),
                       newRotate[0], newRotate[1], newRotate[2], type="float3", edit=1)
            pm.setAttr("{0}.scale".format(obj.replace(instname, dupeOrigName)),
                       newScale[0], newScale[1], newScale[2], type="float3", edit=1)
            pm.setAttr("{0}.visibility".format(obj.replace(
                instname, dupeOrigName)), pm.getAttr("{0}.visibility".format(obj)))
            jWalkDownR(obj, instname, dupeOrigName)


def jWalkDownR(instance, instname, dupeOrigName):
    # same function but for recurtion
    for obj in pm.listRelatives(instance, c=1, f=1):
        if obj.split("|")[-1].find("_j_") > 0 and pm.objExists(obj.split("|")[-1].replace(instname, dupeOrigName)):
            # print obj
            # pm.matchTransform(obj.replace(instname, dupeOrigName),obj,rot=1,pos=1)
            newTranslate = pm.getAttr("{0}.translate".format(obj))[0]
            newRotate = pm.getAttr("{0}.rotate".format(obj))[0]
            newScale = pm.getAttr("{0}.scale".format(obj))[0]
            pm.setAttr("{0}.translate".format(obj.replace(instname, dupeOrigName)),
                       newTranslate[0], newTranslate[1], newTranslate[2], type="float3", edit=1)
            pm.setAttr("{0}.rotate".format(obj.replace(instname, dupeOrigName)),
                       newRotate[0], newRotate[1], newRotate[2], type="float3", edit=1)
            pm.setAttr("{0}.scale".format(obj.replace(instname, dupeOrigName)),
                       newScale[0], newScale[1], newScale[2], type="float3", edit=1)
            pm.setAttr("{0}.visibility".format(obj.replace(
                instname, dupeOrigName)), pm.getAttr("{0}.visibility".format(obj)))
            jWalkDownR(obj, instname, dupeOrigName)


def find_original(instance):
    inst_str = str(instance)
    orig_str = inst_str[0:inst_str.find("_inst")] + "_orig"

    orig = pm.ls(orig_str)
    if orig:
        return orig[0]
    else:
        return None


def replaceOriginal(instance_to_take, orig=None):
    # replaces original object with instance you selected
    pm.undoInfo(openChunk=True)
    if not orig:
        orig = instance_to_take[0:instance_to_take.find("_inst")] + "_orig"
        try:
            orig = pm.ls(orig)[0]
        except IndexError:
            pm.warning("No original object found")
            return
    new_orig = pm.duplicate(instance_to_take, rr=1, un=1)[0]
    old_layer = pm.listConnections(orig, type="displayLayer")
    if old_layer:
        old_layer = old_layer[0]
    else:
        old_layer = "defaultLayer"
    pm.editDisplayLayerMembers(old_layer, new_orig, nr=1)
    parent = orig.getParent()
    pm.matchTransform(new_orig, orig)
    pm.delete(orig)
    pm.rename(new_orig, orig)
    pm.parent(new_orig, parent)
    reset_rig_state(new_orig)

    pm.undoInfo(closeChunk=True)


def delInstances(oldsel):
    # deletes all object instances
    oldSel = pm.ls(sl=1, l=1)
    cleanname = oldSel[0].split("|")[-1].replace("_orig", "")
    for obj in pm.editDisplayLayerMembers("Instances", q=1, fn=1):
        if obj.split("|")[-1].startswith(cleanname):
            parGrp = pm.listRelatives(obj, p=1, f=1)
            if parGrp[0].find("_right") < 0:
                pm.delete(obj)


def create_instances_layer():
   # Create a new layer named "Instances"
    if not pm.objExists("Instances"):
        pm.createDisplayLayer(name="Instances", number=1, empty=1)
        # pm.editDisplayLayerMembers("Instances", "defaultLayer", nr=1)
        # pm.editDisplayLayerGlobals(currentLayer="Instances")


def replace_instances(original):
    print("making instances")

    def replace_instance(original, old_instance):
        print("making instance")
        parGrp = pm.listRelatives(old_instance, p=1, f=1)
        old_index = old_instance.split("_")[-1]
        if not parGrp:
            raise Exception("Object has no parent group")

        if parGrp[0].find("_right") < 0:
            duped = pm.duplicate(original, rr=1, un=1)
            pm.parent(duped[0], parGrp[0])
            newname = original.split(
                "|")[-1].replace("_orig", "_inst_" + old_index)
            pm.matchTransform(duped[0], old_instance)

            pm.editDisplayLayerMembers("Instances", duped[0], nr=1)

            print("Copying rig state from {0} to {1}".format(
                old_instance, duped[0]))
            copyRigState(old_instance, duped[0])

            if old_instance.find("_orig") == -1:
                pm.delete(old_instance)

            pm.rename(duped[0], newname)
        else:
            print("right side")

    # makes instances of object
    cleanname = original.split("|")[-1].replace("orig", "")

    if pm.editDisplayLayerMembers("Instances", q=1, fn=1) is None:
        replace_instance(original)
    else:
        for instance in pm.editDisplayLayerMembers("Instances", q=1, fn=1):
            if instance.split("|")[-1].startswith(cleanname):
                replace_instance(original, instance)


def find_instance_parent(node):
    index = 0
    while index < 10:
        if node.endswith("_orig"):
            return [node, True]

        elif node.split("|")[-1].find("_inst") != -1:
            return [node, False]

        else:
            node = node.getParent()
            if node:
                index += 1
            else:
                break
    return [None, False]


def make_original():
    selection = pm.selected()

    if not pm.objExists("Originals_grp"):
        pm.group(em=1, name="Originals_grp")

    if len(selection) > 1:
        new_orig = pm.group(em=1, name=str(selection[0]) + "_orig")
        pm.parent(new_orig, "Originals_grp")
    else:
        if not selection[0].endswith("_orig"):
            pm.rename(selection[0], str(selection[0]) + "_orig")
        else:
            pm.warning("Object is already an original")
        try:
            pm.parent(selection[0], "Originals_grp")
        except:
            pass


def make_instance(original):
    print("making instance")
    new_instance = pm.duplicate(original, rr=1, un=1, name=str(
        original).replace("_orig", "_inst_0"))[0]

    pm.parent(new_instance, w=1)
    pm.select(new_instance)

    if not pm.objExists("Instances"):
        pm.createDisplayLayer(name="Instances", number=1, empty=1)

    pm.editDisplayLayerMembers("Instances", new_instance, nr=1)


def duplicate_instance(instance):
    pm.duplicate(instance, rr=1, un=1)


def update_instances(instance_parent=[]):
    print("instanciate")
    pm.undoInfo(openChunk=True)
    # Z:\projects\NU\Assets\nopia-produtils\ToolBox\mod_shortCutScripts.py
    create_instances_layer()

    selection = pm.ls(sl=1)

    if instance_parent == []:
        instance_parent = find_instance_parent(selection[0])

    if instance_parent[1]:
        replace_instances(instance_parent[0])

    pm.undoInfo(closeChunk=True)


def selectOriginal():
    sel = pm.ls(sl=1)
    pm.select(sel[0].split("|")[-1]
              [0:sel[0].split("|")[-1].find("inst_")]+"orig")


def export_originals(export_path="C:/_Temp/Instancer/"):
    def export_original(obj):
        # store original object transform
        orig_position = pm.xform(obj, q=1, ws=1, t=1)
        orig_rotation = pm.xform(obj, q=1, ws=1, ro=1)
        orig_scale = pm.xform(obj, q=1, ws=1, s=1)
        # store original object parent
        orig_parent = obj.getParent()
        # Parent object to world and reset transform and freeze transform
        pm.parent(obj, w=1)
        pm.xform(obj, ws=1, t=(0, 0, 0))
        pm.xform(obj, ws=1, ro=(0, 0, 0))
        pm.xform(obj, ws=1, s=(1, 1, 1))

        pm.select(obj, r=1)
        pm.mel.FBXExport(
            f=export_path + str(obj).replace("_orig", "_sm") + ".fbx", s=1)

        # restore original object transform
        pm.xform(obj, ws=1, t=orig_position)
        pm.xform(obj, ws=1, ro=orig_rotation)
        pm.xform(obj, ws=1, s=orig_scale)
        # restore original object parent
        pm.parent(obj, orig_parent)

    selection = pm.ls(sl=1)
    if selection:
        for obj in selection:
            if obj.find("_orig") != -1:
                export_original(obj)
    else:
        if pm.objExists("Originals_grp"):
            for obj in pm.listRelatives("Originals_grp", c=1):
                if obj.endswith("_orig"):
                    pm.select(obj, r=1)
                    export_original(obj)
        else:
            pm.warning("No originals found")
            pm.warning("Create Originals_grp and put originals in it.")


def export_instances(path):
    list_all_instances = pm.editDisplayLayerMembers("Instances", q=1, fn=1)
    list_all_instances = [
        pm.PyNode(x) for x in list_all_instances if x.find("_inst_") != -1]
    write_transforms_to_json(
        path + "instances_transforms_fromMaya.json", list_all_instances)


def update_instances_json(path):
    update_transforms_from_json(path + "instances_transforms_fromUE.json")


##############################################################################
# Exporting to JSON
##############################################################################


def write_transforms_to_json(path, transforms=[]):
    """
    Write transforms to JSON
    """
    if transforms == []:
        transforms = pm.ls(sl=1, type='transform')
    data = {}
    for transform in transforms:
        asset_name = str(transform).split("_")[0]
        id = str(transform).split("_")[-1]
        dict_key = asset_name + "_" + id
        data.update(
            {dict_key:
             [
                 transform.translateX.get(),
                 transform.translateY.get(),
                 transform.translateZ.get(),
                 transform.rotateX.get(),
                 transform.rotateY.get(),
                 transform.rotateZ.get(),
                 transform.scaleX.get(),
                 transform.scaleY.get(),
                 transform.scaleZ.get()
             ]
             })
    # Write to JSON
    with open(path, 'w') as outfile:
        json.dump(data, outfile)


def axis_rotation_to_quaternion(axis, rotation):
    theta = math.radians(rotation) / 2
    w = math.cos(theta)
    x = axis.x * math.sin(theta)
    y = axis.y * math.sin(theta)
    z = axis.z * math.sin(theta)
    return om.MQuaternion(x, y, z, w)


def rotator_to_euler(rotator):
    """
    A function converts rotator to euler angles.
    """
    qx = axis_rotation_to_quaternion(om.MVector(1, 0, 0), rotator[0])
    qy = axis_rotation_to_quaternion(om.MVector(0, 1, 0), -rotator[2])
    qz = axis_rotation_to_quaternion(om.MVector(0, 0, 1), -rotator[1])

    qu = qx * qy * qz
    qu_e = qu.asEulerRotation()
    rad = 57.296
    return [math.degrees(qu_e.x), math.degrees(qu_e.y), math.degrees(qu_e.z)]


def read_json(json_file):
    """
    A function reads a json file and returns a dictionary.
    """
    with open(json_file) as json_file:
        data = json.load(json_file)
    return data


def update_transforms_from_json(json_path):
    json_data = read_json(json_path)

    pm.undoInfo(openChunk=1)
    for key in json_data.keys():
        asset_name = key.split("_")[0]
        id = key.split("_")[1]

        try:
            transform = pm.ls(key)[0]
        except IndexError:
            try:
                transform = pm.ls(asset_name + "_inst_" + id)[0]
            except IndexError:
                # TODO: Should create one from existing
                print("Transform not found: " + key)
                continue

        position = (json_data[key][0][0], json_data[key]
                    [0][2], json_data[key][0][1])
        rotation = (json_data[key][1][0], json_data[key]
                    [1][1], json_data[key][1][2])
        scale = (json_data[key][2][0], json_data[key]
                 [2][2], json_data[key][2][1])

        rotation = rotator_to_euler(rotation)
        pm.xform(transform, ws=1, t=position)
        pm.xform(transform, ws=1, ro=rotation)
        pm.xform(transform, ws=1, s=scale)
    print("Transforms Updated")
    pm.undoInfo(closeChunk=1)


if __name__ == "__main__":
    InstancerMainWindow()

    test_string = "teststring"

    print(test_string.split("_")[-1])
